#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import ast
from copy import deepcopy
from unittest import mock

import pytest
from google.api_core.exceptions import NotFound
from google.api_core.retry import Retry
from google.cloud.compute_v1.types import Instance, InstanceGroupManager, InstanceTemplate

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.cloud.operators.compute import (
    ComputeEngineCopyInstanceTemplateOperator,
    ComputeEngineDeleteInstanceGroupManagerOperator,
    ComputeEngineDeleteInstanceOperator,
    ComputeEngineDeleteInstanceTemplateOperator,
    ComputeEngineInsertInstanceFromTemplateOperator,
    ComputeEngineInsertInstanceGroupManagerOperator,
    ComputeEngineInsertInstanceOperator,
    ComputeEngineInsertInstanceTemplateOperator,
    ComputeEngineInstanceGroupUpdateManagerTemplateOperator,
    ComputeEngineSetMachineTypeOperator,
    ComputeEngineStartInstanceOperator,
    ComputeEngineStopInstanceOperator,
)
from airflow.utils import timezone

EMPTY_CONTENT = b""
GCP_PROJECT_ID = "project-id"
GCE_ZONE = "zone"
GCE_RESOURCE_ID = "resource-id"
GCE_SHORT_MACHINE_TYPE_NAME = "n1-machine-type"
GCE_SOURCE_IMAGE = "projects/debian-cloud/global/images/debian-11-bullseye-v20220621"
GCP_INSTANCE_TEMPLATE_NAME = "instance-template-name"
TASK_ID = "task-id"

SET_MACHINE_TYPE_BODY = {"machineType": f"zones/{GCE_ZONE}/machineTypes/{GCE_SHORT_MACHINE_TYPE_NAME}"}
SOURCE_INSTANCE_TEMPLATE = f"global/InstanceTemplates/{GCP_INSTANCE_TEMPLATE_NAME}"
DEFAULT_DATE = timezone.datetime(2017, 1, 1)

GCP_CONN_ID = "test-conn"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]
TIMEOUT = 120
RETRY = mock.MagicMock(Retry)
METADATA = [("key", "value")]
API_VERSION = "v1"

COMPUTE_ENGINE_HOOK_PATH = "airflow.providers.google.cloud.operators.compute.ComputeEngineHook"
IGM_PATH = "airflow.providers.google.cloud.operators.compute.InstanceGroupManager"

GCP_INSTANCE_BODY_FROM_TEMPLATE = {"name": "a-instance-from-template"}
GCE_INSTANCE_BODY_API_CALL = {
    "name": "instance-name",
    "machine_type": f"zones/{GCE_ZONE}/machineTypes/{GCE_SHORT_MACHINE_TYPE_NAME}",
    "disks": [
        {
            "boot": True,
            "device_name": "instance-name",
            "initialize_params": {
                "disk_size_gb": "10",
                "disk_type": f"zones/{GCE_ZONE}/diskTypes/pd-balanced",
                "source_image": GCE_SOURCE_IMAGE,
            },
        }
    ],
    "network_interfaces": [
        {
            "access_configs": [{"name": "External NAT", "network_tier": "PREMIUM"}],
            "stack_type": "IPV4_ONLY",
            "subnetwork": "regions/us-central1/subnetworks/default",
        }
    ],
}
GCE_INSTANCE_BODY_WITHOUT_NAME_API_CALL = {
    "machine_type": f"zones/{GCE_ZONE}/machineTypes/{GCE_SHORT_MACHINE_TYPE_NAME}",
    "disks": [
        {
            "boot": True,
            "device_name": "instance-name",
            "initialize_params": {
                "disk_size_gb": "10",
                "disk_type": f"zones/{GCE_ZONE}/diskTypes/pd-balanced",
                "source_image": GCE_SOURCE_IMAGE,
            },
        }
    ],
    "network_interfaces": [
        {
            "access_configs": [{"name": "External NAT", "network_tier": "PREMIUM"}],
            "stack_type": "IPV4_ONLY",
            "subnetwork": "regions/us-central1/subnetworks/default",
        }
    ],
}


class TestGceInstanceInsert:
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_insert_instance_should_execute_successfully(self, mock_hook):
        get_instance_obj_mock = mock.MagicMock()
        get_instance_obj_mock.__class__ = Instance
        mock_hook.return_value.get_instance.side_effect = [
            NotFound("Error message"),
            get_instance_obj_mock,
        ]
        op = ComputeEngineInsertInstanceOperator(
            project_id=GCP_PROJECT_ID,
            resource_id=GCE_RESOURCE_ID,
            body=GCE_INSTANCE_BODY_API_CALL,
            zone=GCE_ZONE,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            api_version=API_VERSION,
        )
        mock_hook.return_value.insert_instance.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            body=GCE_INSTANCE_BODY_API_CALL,
            zone=GCE_ZONE,
            request_id=None,
        )

    def test_insert_instance_should_throw_ex_when_missing_project_id(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'project_id' is missing"):
            ComputeEngineInsertInstanceOperator(
                project_id="",
                body=GCE_INSTANCE_BODY_API_CALL,
                zone=GCE_ZONE,
                task_id=TASK_ID,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_insert_instance_should_not_throw_ex_when_project_id_none(self, mock_hook):
        get_instance_obj_mock = mock.MagicMock()
        get_instance_obj_mock.__class__ = Instance
        mock_hook.return_value.get_instance.side_effect = [
            NotFound("Error message"),
            get_instance_obj_mock,
        ]
        op = ComputeEngineInsertInstanceOperator(
            resource_id=GCE_RESOURCE_ID,
            body=GCE_INSTANCE_BODY_API_CALL,
            zone=GCE_ZONE,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance.assert_called_once_with(
            body=GCE_INSTANCE_BODY_API_CALL,
            zone=GCE_ZONE,
            request_id=None,
            project_id=None,
        )

    def test_insert_instance_should_throw_ex_when_missing_zone(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'zone' is missing"):
            ComputeEngineInsertInstanceOperator(
                resource_id=GCE_RESOURCE_ID,
                body=GCE_INSTANCE_BODY_API_CALL,
                zone="",
                task_id=TASK_ID,
                retry=RETRY,
                timeout=TIMEOUT,
                metadata=METADATA,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    def test_insert_instance_should_throw_ex_when_missing_resource_id(self):
        with pytest.raises(
            AirflowException,
            match=r"The required parameters 'resource_id' and "
            r"body\['name'\] are missing\. Please, provide "
            r"at least one of them",
        ):
            ComputeEngineInsertInstanceOperator(
                project_id=GCP_PROJECT_ID,
                zone=GCE_ZONE,
                body=GCE_INSTANCE_BODY_WITHOUT_NAME_API_CALL,
                task_id=TASK_ID,
                resource_id="",
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_insert_instance_should_not_throw_ex_when_name_is_templated(self, mock_hook):
        get_instance_obj_mock = mock.MagicMock()
        get_instance_obj_mock.__class__ = Instance
        mock_hook.return_value.get_instance.side_effect = [
            NotFound("Error message"),
            get_instance_obj_mock,
        ]
        body_with_templated_name = deepcopy(GCE_INSTANCE_BODY_API_CALL)
        body_with_templated_name["name"] = "{{ logical_date }}"
        op = ComputeEngineInsertInstanceOperator(
            project_id=GCP_PROJECT_ID,
            resource_id=GCE_RESOURCE_ID,
            body=body_with_templated_name,
            zone=GCE_ZONE,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            body=body_with_templated_name,
            zone=GCE_ZONE,
            request_id=None,
        )


class TestGceInstanceInsertFromTemplate:
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_insert_instance_from_template_should_execute_successfully(self, mock_hook):
        get_instance_obj_mock = mock.MagicMock()
        get_instance_obj_mock.__class__ = Instance
        mock_hook.return_value.get_instance.side_effect = [
            NotFound("Error message"),
            get_instance_obj_mock,
        ]
        op = ComputeEngineInsertInstanceFromTemplateOperator(
            project_id=GCP_PROJECT_ID,
            source_instance_template=SOURCE_INSTANCE_TEMPLATE,
            body=GCP_INSTANCE_BODY_FROM_TEMPLATE,
            zone=GCE_ZONE,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            body=GCP_INSTANCE_BODY_FROM_TEMPLATE,
            source_instance_template=SOURCE_INSTANCE_TEMPLATE,
            request_id=None,
        )

    def test_insert_instance_from_template_should_throw_ex_when_missing_project_id(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'project_id' is missing"):
            ComputeEngineInsertInstanceFromTemplateOperator(
                project_id="",
                source_instance_template=SOURCE_INSTANCE_TEMPLATE,
                body=GCP_INSTANCE_BODY_FROM_TEMPLATE,
                zone=GCE_ZONE,
                task_id=TASK_ID,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_insert_instance_from_template_should_not_throw_ex_when_project_id_none(self, mock_hook):
        get_instance_obj_mock = mock.MagicMock()
        get_instance_obj_mock.__class__ = Instance
        mock_hook.return_value.get_instance.side_effect = [
            NotFound("Error message"),
            get_instance_obj_mock,
        ]
        op = ComputeEngineInsertInstanceFromTemplateOperator(
            source_instance_template=SOURCE_INSTANCE_TEMPLATE,
            body=GCP_INSTANCE_BODY_FROM_TEMPLATE,
            zone=GCE_ZONE,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance.assert_called_once_with(
            project_id=None,
            zone=GCE_ZONE,
            body=GCP_INSTANCE_BODY_FROM_TEMPLATE,
            source_instance_template=SOURCE_INSTANCE_TEMPLATE,
            request_id=None,
        )

    def test_insert_instance_from_template_should_throw_ex_when_missing_zone(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'zone' is missing"):
            ComputeEngineInsertInstanceFromTemplateOperator(
                project_id=GCP_PROJECT_ID,
                zone="",
                source_instance_template=SOURCE_INSTANCE_TEMPLATE,
                body=GCP_INSTANCE_BODY_FROM_TEMPLATE,
                task_id=TASK_ID,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    def test_insert_instance_from_template_should_throw_ex_when_missing_source_instance_template(self):
        with pytest.raises(
            (TypeError, AirflowException), match=r"missing keyword argument 'source_instance_template'"
        ):
            ComputeEngineInsertInstanceFromTemplateOperator(
                project_id=GCP_PROJECT_ID,
                body=GCP_INSTANCE_BODY_FROM_TEMPLATE,
                zone=GCE_ZONE,
                task_id=TASK_ID,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    def test_insert_instance_from_template_should_throw_ex_when_missing_body(self):
        with pytest.raises((TypeError, AirflowException), match=r"missing keyword argument 'body'"):
            ComputeEngineInsertInstanceFromTemplateOperator(
                project_id=GCP_PROJECT_ID,
                source_instance_template=SOURCE_INSTANCE_TEMPLATE,
                zone=GCE_ZONE,
                task_id=TASK_ID,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_insert_instance_from_template_should_not_throw_ex_when_name_is_templated(self, mock_hook):
        get_instance_obj_mock = mock.MagicMock()
        get_instance_obj_mock.__class__ = Instance
        mock_hook.return_value.get_instance.side_effect = [
            NotFound("Error message"),
            get_instance_obj_mock,
        ]
        body_with_templated_name = deepcopy(GCP_INSTANCE_BODY_FROM_TEMPLATE)
        body_with_templated_name["name"] = "{{ execution_date }}"
        op = ComputeEngineInsertInstanceFromTemplateOperator(
            project_id=GCP_PROJECT_ID,
            source_instance_template=SOURCE_INSTANCE_TEMPLATE,
            body=body_with_templated_name,
            zone=GCE_ZONE,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            body=body_with_templated_name,
            source_instance_template=SOURCE_INSTANCE_TEMPLATE,
            request_id=None,
        )


class TestGceInstanceDelete:
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_delete_instance_should_execute_successfully(self, mock_hook):
        op = ComputeEngineDeleteInstanceOperator(
            resource_id=GCE_RESOURCE_ID,
            zone=GCE_ZONE,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_instance.assert_called_once_with(
            project_id=None,
            request_id=None,
            resource_id=GCE_RESOURCE_ID,
            zone=GCE_ZONE,
        )

    def test_delete_instance_should_throw_ex_when_missing_zone(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'zone' is missing"):
            ComputeEngineDeleteInstanceOperator(
                resource_id=GCE_RESOURCE_ID,
                zone="",
                task_id=TASK_ID,
                retry=RETRY,
                timeout=TIMEOUT,
                metadata=METADATA,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    def test_delete_instance_should_throw_ex_when_missing_resource_id(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'resource_id' is missing"):
            ComputeEngineDeleteInstanceOperator(
                resource_id="",
                zone=GCE_ZONE,
                task_id=TASK_ID,
                retry=RETRY,
                timeout=TIMEOUT,
                metadata=METADATA,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )


class TestGceInstanceStart:
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_start_instance_should_execute_successfully(self, mock_hook):
        mock_hook.return_value.start_instance.return_value = True
        op = ComputeEngineStartInstanceOperator(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            resource_id=GCE_RESOURCE_ID,
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        op.execute(context=context)

        mock_hook.assert_called_once_with(
            api_version="v1",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.start_instance.assert_called_once_with(
            zone=GCE_ZONE, resource_id=GCE_RESOURCE_ID, project_id=GCP_PROJECT_ID
        )

    # Setting all the operator's input parameters as template dag_ids
    # (could be anything else) just to test if the templating works for all fields
    @pytest.mark.db_test
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_start_instance_with_templates(self, _, create_task_instance_of_operator, session):
        dag_id = "test_instance_start_with_templates"
        ti = create_task_instance_of_operator(
            ComputeEngineStartInstanceOperator,
            dag_id=dag_id,
            project_id="{{ dag.dag_id }}",
            zone="{{ dag.dag_id }}",
            resource_id="{{ dag.dag_id }}",
            gcp_conn_id="{{ dag.dag_id }}",
            api_version="{{ dag.dag_id }}",
            task_id="id",
        )
        rendered = ti.render_templates()
        assert dag_id == rendered.project_id
        assert dag_id == rendered.zone
        assert dag_id == rendered.resource_id
        assert dag_id == rendered.gcp_conn_id
        assert dag_id == rendered.api_version

    def test_instance_start_should_throw_ex_when_missing_project_id(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'project_id' is missing"):
            ComputeEngineStartInstanceOperator(
                project_id="",
                zone=GCE_ZONE,
                resource_id=GCE_RESOURCE_ID,
                task_id=TASK_ID,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_instance_start_should_not_throw_ex_when_project_id_none(self, _):
        op = ComputeEngineStartInstanceOperator(
            zone=GCE_ZONE,
            resource_id=GCE_RESOURCE_ID,
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        op.execute(context=context)

    def test_instance_start_should_throw_ex_when_missing_zone(self):
        with pytest.raises(AirflowException, match=r"he required parameter 'zone' is missing"):
            ComputeEngineStartInstanceOperator(
                project_id=GCP_PROJECT_ID,
                zone="",
                resource_id=GCE_RESOURCE_ID,
                task_id=TASK_ID,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    def test_instance_start_should_throw_ex_when_missing_resource_id(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'resource_id' is missing"):
            ComputeEngineStartInstanceOperator(
                project_id=GCP_PROJECT_ID,
                zone=GCE_ZONE,
                resource_id="",
                task_id=TASK_ID,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )


class TestGceInstanceStop:
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_instance_stop_should_execute_successfully(self, mock_hook):
        op = ComputeEngineStopInstanceOperator(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            resource_id=GCE_RESOURCE_ID,
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            api_version="v1",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.stop_instance.assert_called_once_with(
            zone=GCE_ZONE, resource_id=GCE_RESOURCE_ID, project_id=GCP_PROJECT_ID
        )

    # Setting all the operator's input parameters as templated dag_ids
    # (could be anything else) just to test if the templating works for all fields
    @pytest.mark.db_test
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_instance_stop_with_templates(self, _, create_task_instance_of_operator, session):
        dag_id = "test_instance_stop_with_templates"
        ti = create_task_instance_of_operator(
            ComputeEngineStopInstanceOperator,
            dag_id=dag_id,
            project_id="{{ dag.dag_id }}",
            zone="{{ dag.dag_id }}",
            resource_id="{{ dag.dag_id }}",
            gcp_conn_id="{{ dag.dag_id }}",
            api_version="{{ dag.dag_id }}",
            task_id="id",
        )
        rendered = ti.render_templates()
        assert dag_id == rendered.project_id
        assert dag_id == rendered.zone
        assert dag_id == rendered.resource_id
        assert dag_id == rendered.gcp_conn_id
        assert dag_id == rendered.api_version

    def test_instance_stop_should_throw_ex_when_missing_project_id(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'project_id' is missing"):
            ComputeEngineStopInstanceOperator(
                project_id="", zone=GCE_ZONE, resource_id=GCE_RESOURCE_ID, task_id="id"
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_instance_stop_should_not_throw_ex_when_project_id_none(self, mock_hook):
        op = ComputeEngineStopInstanceOperator(zone=GCE_ZONE, resource_id=GCE_RESOURCE_ID, task_id="id")
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            api_version="v1",
            gcp_conn_id="google_cloud_default",
            impersonation_chain=None,
        )
        mock_hook.return_value.stop_instance.assert_called_once_with(
            zone=GCE_ZONE, resource_id=GCE_RESOURCE_ID, project_id=None
        )

    def test_instance_stop_should_throw_ex_when_missing_zone(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'zone' is missing"):
            ComputeEngineStopInstanceOperator(
                project_id=GCP_PROJECT_ID, zone="", resource_id=GCE_RESOURCE_ID, task_id="id"
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_instance_stop_should_throw_ex_when_missing_resource_id(self, mock_hook):
        with pytest.raises(AirflowException, match=r"The required parameter 'resource_id' is missing"):
            ComputeEngineStopInstanceOperator(
                project_id=GCP_PROJECT_ID, zone=GCE_ZONE, resource_id="", task_id="id"
            )


class TestGceInstanceSetMachineType:
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_machine_type_set_should_execute_successfully(self, mock_hook):
        mock_hook.return_value.set_machine_type.return_value = True
        op = ComputeEngineSetMachineTypeOperator(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            resource_id=GCE_RESOURCE_ID,
            body=SET_MACHINE_TYPE_BODY,
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            api_version="v1",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.set_machine_type.assert_called_once_with(
            zone=GCE_ZONE, resource_id=GCE_RESOURCE_ID, body=SET_MACHINE_TYPE_BODY, project_id=GCP_PROJECT_ID
        )

    # Setting all the operator's input parameters as templated dag_ids
    # (could be anything else) just to test if the templating works for all fields
    @pytest.mark.db_test
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_machine_type_set_with_templates(self, _, create_task_instance_of_operator, session):
        dag_id = "test_set_machine_type_with_templates"
        ti = create_task_instance_of_operator(
            ComputeEngineSetMachineTypeOperator,
            dag_id=dag_id,
            project_id="{{ dag.dag_id }}",
            zone="{{ dag.dag_id }}",
            resource_id="{{ dag.dag_id }}",
            body={},
            gcp_conn_id="{{ dag.dag_id }}",
            api_version="{{ dag.dag_id }}",
            task_id="id",
        )
        rendered = ti.render_templates()
        assert dag_id == rendered.project_id
        assert dag_id == rendered.zone
        assert dag_id == rendered.resource_id
        assert dag_id == rendered.gcp_conn_id
        assert dag_id == rendered.api_version

    def test_machine_type_set_should_throw_ex_when_missing_project_id(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'project_id' is missing"):
            ComputeEngineSetMachineTypeOperator(
                project_id="",
                zone=GCE_ZONE,
                resource_id=GCE_RESOURCE_ID,
                body=SET_MACHINE_TYPE_BODY,
                task_id=TASK_ID,
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_machine_type_set_should_not_throw_ex_when_project_id_none(self, mock_hook):
        op = ComputeEngineSetMachineTypeOperator(
            zone=GCE_ZONE,
            resource_id=GCE_RESOURCE_ID,
            body=SET_MACHINE_TYPE_BODY,
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            api_version="v1",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.set_machine_type.assert_called_once_with(
            zone=GCE_ZONE, resource_id=GCE_RESOURCE_ID, body=SET_MACHINE_TYPE_BODY, project_id=None
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_machine_type_set_should_throw_ex_when_missing_zone(self, mock_hook):
        with pytest.raises(AirflowException, match=r"The required parameter 'zone' is missing"):
            ComputeEngineSetMachineTypeOperator(
                project_id=GCP_PROJECT_ID,
                zone="",
                resource_id=GCE_RESOURCE_ID,
                body=SET_MACHINE_TYPE_BODY,
                task_id=TASK_ID,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    def test_machine_type_set_should_throw_ex_when_missing_resource_id(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'resource_id' is missing"):
            ComputeEngineSetMachineTypeOperator(
                project_id=GCP_PROJECT_ID,
                zone=GCE_ZONE,
                resource_id="",
                body=SET_MACHINE_TYPE_BODY,
                task_id=TASK_ID,
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_machine_type_set_should_throw_ex_when_missing_machine_type(self, mock_hook):
        op = ComputeEngineSetMachineTypeOperator(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            resource_id=GCE_RESOURCE_ID,
            body={},
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        with pytest.raises(AirflowException, match=r"The required body field 'machineType' is missing"):
            op.execute(context=context)
        mock_hook.assert_called_once_with(
            api_version="v1",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

    MOCK_OP_RESPONSE = (
        "{'kind': 'compute#operation', 'id': '8529919847974922736', "
        "'name': "
        "'operation-1538578207537-577542784f769-7999ab71-94f9ec1d', "
        "'zone': 'https://www.googleapis.com/compute/v1/projects/example"
        "-project/zones/europe-west3-b', 'operationType': "
        "'setMachineType', 'targetLink': "
        "'https://www.googleapis.com/compute/v1/projects/example-project"
        "/zones/europe-west3-b/instances/pa-1', 'targetId': "
        "'2480086944131075860', 'status': 'DONE', 'user': "
        "'service-account@example-project.iam.gserviceaccount.com', "
        "'progress': 100, 'insertTime': '2018-10-03T07:50:07.951-07:00', "
        "'startTime': '2018-10-03T07:50:08.324-07:00', 'endTime': "
        "'2018-10-03T07:50:08.484-07:00', 'error': {'errors': [{'code': "
        "'UNSUPPORTED_OPERATION', 'message': \"Machine type with name "
        "'machine-type-1' does not exist in zone 'europe-west3-b'.\"}]}, "
        "'httpErrorStatusCode': 400, 'httpErrorMessage': 'BAD REQUEST', "
        "'selfLink': "
        "'https://www.googleapis.com/compute/v1/projects/example-project"
        "/zones/europe-west3-b/operations/operation-1538578207537"
        "-577542784f769-7999ab71-94f9ec1d'} "
    )

    @pytest.mark.db_test
    @mock.patch(
        "airflow.providers.google.cloud.operators.compute.ComputeEngineHook._check_zone_operation_status"
    )
    @mock.patch(
        "airflow.providers.google.cloud.operators.compute.ComputeEngineHook._execute_set_machine_type"
    )
    @mock.patch("airflow.providers.google.cloud.operators.compute.ComputeEngineHook.get_conn")
    def test_machine_type_set_should_handle_and_trim_gce_error(
        self, get_conn, _execute_set_machine_type, _check_zone_operation_status
    ):
        get_conn.return_value = {}
        _execute_set_machine_type.return_value = {"name": "test-operation"}
        _check_zone_operation_status.return_value = ast.literal_eval(self.MOCK_OP_RESPONSE)
        op = ComputeEngineSetMachineTypeOperator(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            resource_id=GCE_RESOURCE_ID,
            body=SET_MACHINE_TYPE_BODY,
            task_id=TASK_ID,
        )
        context = mock.MagicMock()
        with pytest.raises(AirflowException, match=r"400 BAD REQUEST: {.+UNSUPPORTED_OPERATION"):
            op.execute(context=context)
        _check_zone_operation_status.assert_called_once_with(
            {}, "test-operation", GCP_PROJECT_ID, GCE_ZONE, mock.ANY
        )
        _execute_set_machine_type.assert_called_once_with(
            GCE_ZONE, GCE_RESOURCE_ID, SET_MACHINE_TYPE_BODY, GCP_PROJECT_ID
        )


GCE_INSTANCE_TEMPLATE_NAME_API_CALL = "instance-template-test-api"
GCE_INSTANCE_TEMPLATE_BODY_API_CALL = {
    "name": GCE_INSTANCE_TEMPLATE_NAME_API_CALL,
    "properties": {
        "machine_type": GCE_SHORT_MACHINE_TYPE_NAME,
        "disks": [
            {
                "auto_delete": True,
                "boot": True,
                "device_name": GCE_INSTANCE_TEMPLATE_NAME_API_CALL,
                "initialize_params": {
                    "disk_size_gb": "10",
                    "disk_type": "pd-balanced",
                    "source_image": GCE_SOURCE_IMAGE,
                },
            }
        ],
        "network_interfaces": [{"network": "/global/networks/default"}],
    },
}


class TestGceTemplateInsert:
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_insert_template_should_execute_successfully(self, mock_hook):
        get_template_obj_mock = mock.MagicMock()
        get_template_obj_mock.__class__ = InstanceTemplate
        mock_hook.return_value.get_instance_template.side_effect = [
            NotFound("Error message"),
            get_template_obj_mock,
        ]
        op = ComputeEngineInsertInstanceTemplateOperator(
            project_id=GCP_PROJECT_ID,
            body=GCE_INSTANCE_TEMPLATE_BODY_API_CALL,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance_template.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            body=GCE_INSTANCE_TEMPLATE_BODY_API_CALL,
            request_id=None,
        )

    def test_insert_template_should_throw_ex_when_missing_project_id(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'project_id' is missing"):
            ComputeEngineInsertInstanceTemplateOperator(
                project_id="",
                body=GCE_INSTANCE_TEMPLATE_BODY_API_CALL,
                task_id=TASK_ID,
                retry=RETRY,
                timeout=TIMEOUT,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_insert_template_should_not_throw_ex_when_project_id_none(self, mock_hook):
        get_template_obj_mock = mock.MagicMock()
        get_template_obj_mock.__class__ = InstanceTemplate
        mock_hook.return_value.get_instance_template.side_effect = [
            NotFound("Error message"),
            get_template_obj_mock,
        ]
        op = ComputeEngineInsertInstanceTemplateOperator(
            body=GCE_INSTANCE_TEMPLATE_BODY_API_CALL,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance_template.assert_called_once_with(
            project_id=None,
            body=GCE_INSTANCE_TEMPLATE_BODY_API_CALL,
            request_id=None,
        )

    def test_insert_template_should_throw_ex_when_missing_body(self):
        with pytest.raises((TypeError, AirflowException), match=r"missing keyword argument 'body'"):
            ComputeEngineInsertInstanceTemplateOperator(
                task_id=TASK_ID,
                project_id=GCP_PROJECT_ID,
                retry=RETRY,
                timeout=TIMEOUT,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_insert_template_should_not_throw_ex_when_name_is_templated(self, mock_hook):
        get_template_obj_mock = mock.MagicMock()
        get_template_obj_mock.__class__ = InstanceTemplate
        mock_hook.return_value.get_instance_template.side_effect = [
            NotFound("Error message"),
            get_template_obj_mock,
        ]
        body_with_templated_name = deepcopy(GCE_INSTANCE_TEMPLATE_BODY_API_CALL)
        body_with_templated_name["name"] = "{{ execution_date }}"
        op = ComputeEngineInsertInstanceTemplateOperator(
            project_id=GCP_PROJECT_ID,
            body=body_with_templated_name,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance_template.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            body=body_with_templated_name,
            request_id=None,
        )


class TestGceTemplateDelete:
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_delete_template_should_execute_successfully(self, mock_hook):
        op = ComputeEngineDeleteInstanceTemplateOperator(
            resource_id=GCE_RESOURCE_ID,
            project_id=GCP_PROJECT_ID,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_instance_template.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            request_id=None,
            resource_id=GCE_RESOURCE_ID,
        )

    def test_delete_template_should_throw_ex_when_missing_project_id(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'project_id' is missing"):
            ComputeEngineDeleteInstanceTemplateOperator(
                project_id="",
                resource_id=GCE_RESOURCE_ID,
                task_id=TASK_ID,
                retry=RETRY,
                timeout=TIMEOUT,
                metadata=METADATA,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_delete_template_should_not_throw_ex_when_project_id_none(self, mock_hook):
        op = ComputeEngineDeleteInstanceTemplateOperator(
            resource_id=GCE_RESOURCE_ID,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_instance_template.assert_called_once_with(
            resource_id=GCE_RESOURCE_ID,
            project_id=None,
            request_id=None,
        )

    def test_delete_template_should_throw_ex_when_missing_resource_id(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'resource_id' is missing"):
            ComputeEngineDeleteInstanceTemplateOperator(
                resource_id="",
                project_id=GCP_PROJECT_ID,
                task_id=TASK_ID,
                retry=RETRY,
                timeout=TIMEOUT,
                metadata=METADATA,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )


GCE_INSTANCE_TEMPLATE_NAME = "instance-template-test"
GCE_INSTANCE_TEMPLATE_NEW_NAME = "instance-template-test-new"
GCE_INSTANCE_TEMPLATE_REQUEST_ID = "e12d5b48-4826-4ba9-ada6-0cff1e0b36a6"
GCE_INSTANCE_TEMPLATE_BODY_GET = {
    "kind": "compute#instanceTemplate",
    "id": "6950321349997439715",
    "creation_timestamp": "2018-10-15T06:20:12.777-07:00",
    "name": GCE_INSTANCE_TEMPLATE_NAME,
    "description": "",
    "properties": {
        "machine_type": "n1-standard-1",
        "network_interfaces": [
            {
                "kind": "compute#networkInterface",
                "network": "https://www.googleapis.com/compute/v1/projects/project/global/networks/default",
                "access_configs": [
                    {
                        "kind": "compute#accessConfig",
                        "type": "ONE_TO_ONE_NAT",
                    }
                ],
            },
            {
                "network": "https://www.googleapis.com/compute/v1/projects/project/global/networks/default",
                "access_configs": [{"kind": "compute#accessConfig", "networkTier": "PREMIUM"}],
            },
        ],
        "disks": [
            {
                "kind": "compute#attachedDisk",
                "type": "PERSISTENT",
                "licenses": [
                    "A String",
                ],
            }
        ],
    },
    "self_link": "https://www.googleapis.com/compute/v1/projects/project"
    "/global/instanceTemplates/instance-template-test",
}
GCE_INSTANCE_TEMPLATE_BODY_INSERT = {
    "name": GCE_INSTANCE_TEMPLATE_NEW_NAME,
}
GCE_INSTANCE_TEMPLATE_BODY_GET_NEW = deepcopy(GCE_INSTANCE_TEMPLATE_BODY_GET)
GCE_INSTANCE_TEMPLATE_BODY_GET_NEW["name"] = GCE_INSTANCE_TEMPLATE_NEW_NAME


class TestGceInstanceTemplateCopy:
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_copy_template_should_execute_successfully(self, mock_hook):
        get_template_obj_mock = mock.MagicMock()
        get_template_obj_mock.__class__ = InstanceTemplate
        mock_hook.return_value.get_instance_template.side_effect = [
            NotFound("Error message"),
            get_template_obj_mock,
            get_template_obj_mock,
        ]
        op = ComputeEngineCopyInstanceTemplateOperator(
            project_id=GCP_PROJECT_ID,
            resource_id=GCE_INSTANCE_TEMPLATE_NAME,
            task_id=TASK_ID,
            body_patch=GCE_INSTANCE_TEMPLATE_BODY_INSERT,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            api_version="v1",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance_template.assert_called_once_with(
            project_id=GCP_PROJECT_ID, body=GCE_INSTANCE_TEMPLATE_BODY_INSERT, request_id=None
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_copy_template_should_not_throw_ex_when_missing_project_id(self, mock_hook):
        get_template_obj_mock = mock.MagicMock()
        get_template_obj_mock.__class__ = InstanceTemplate
        mock_hook.return_value.get_instance_template.side_effect = [
            NotFound("Error message"),
            get_template_obj_mock,
            get_template_obj_mock,
        ]
        op = ComputeEngineCopyInstanceTemplateOperator(
            resource_id=GCE_INSTANCE_TEMPLATE_NAME,
            task_id=TASK_ID,
            body_patch=GCE_INSTANCE_TEMPLATE_BODY_INSERT,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            api_version="v1",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance_template.assert_called_once_with(
            project_id=None, body=GCE_INSTANCE_TEMPLATE_BODY_INSERT, request_id=None
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_idempotent_copy_template_when_already_copied_should_execute_successfully(self, mock_hook):
        get_template_obj_mock = mock.MagicMock()
        get_template_obj_mock.__class__ = InstanceTemplate
        mock_hook.return_value.get_instance_template.side_effect = [
            get_template_obj_mock,
            get_template_obj_mock,
            get_template_obj_mock,
        ]
        op = ComputeEngineCopyInstanceTemplateOperator(
            project_id=GCP_PROJECT_ID,
            resource_id=GCE_INSTANCE_TEMPLATE_NAME,
            task_id=TASK_ID,
            body_patch={"name": GCE_INSTANCE_TEMPLATE_NEW_NAME},
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            api_version="v1",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance_template.assert_not_called()

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_copy_template_with_request_id_should_execute_successfully(self, mock_hook):
        get_template_obj_mock = mock.MagicMock()
        get_template_obj_mock.__class__ = InstanceTemplate
        mock_hook.return_value.get_instance_template.side_effect = [
            NotFound("Error message"),
            get_template_obj_mock,
            get_template_obj_mock,
        ]
        op = ComputeEngineCopyInstanceTemplateOperator(
            project_id=GCP_PROJECT_ID,
            resource_id=GCE_INSTANCE_TEMPLATE_NAME,
            request_id=GCE_INSTANCE_TEMPLATE_REQUEST_ID,
            task_id=TASK_ID,
            body_patch=GCE_INSTANCE_TEMPLATE_BODY_INSERT,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            api_version="v1",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance_template.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            body=GCE_INSTANCE_TEMPLATE_BODY_INSERT,
            request_id=GCE_INSTANCE_TEMPLATE_REQUEST_ID,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_copy_template_with_description_fields_should_execute_successfully(self, mock_hook):
        get_template_obj_mock = mock.MagicMock()
        get_template_obj_mock.__class__ = InstanceTemplate
        mock_hook.return_value.get_instance_template.side_effect = [
            NotFound("Error message"),
            get_template_obj_mock,
            get_template_obj_mock,
        ]
        op = ComputeEngineCopyInstanceTemplateOperator(
            project_id=GCP_PROJECT_ID,
            resource_id=GCE_INSTANCE_TEMPLATE_NAME,
            request_id=GCE_INSTANCE_TEMPLATE_REQUEST_ID,
            task_id=TASK_ID,
            body_patch={"name": GCE_INSTANCE_TEMPLATE_NEW_NAME, "description": "New description"},
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            api_version="v1",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        body_insert = deepcopy(GCE_INSTANCE_TEMPLATE_BODY_INSERT)
        body_insert["description"] = "New description"
        mock_hook.return_value.insert_instance_template.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            body=body_insert,
            request_id=GCE_INSTANCE_TEMPLATE_REQUEST_ID,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_copy_template_with_some_validation_warnings_should_execute_successfully(self, mock_hook):
        get_template_obj_mock = mock.MagicMock()
        get_template_obj_mock.__class__ = InstanceTemplate
        mock_hook.return_value.get_instance_template.side_effect = [
            NotFound("Error message"),
            get_template_obj_mock,
            get_template_obj_mock,
        ]
        op = ComputeEngineCopyInstanceTemplateOperator(
            project_id=GCP_PROJECT_ID,
            resource_id=GCE_INSTANCE_TEMPLATE_NAME,
            task_id=TASK_ID,
            body_patch={
                "name": GCE_INSTANCE_TEMPLATE_NEW_NAME,
                "some_wrong_field": "test",
                "properties": {"some_other_wrong_field": "test"},
            },
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            api_version="v1",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        body_insert = deepcopy(GCE_INSTANCE_TEMPLATE_BODY_INSERT)
        body_insert["some_wrong_field"] = "test"
        body_insert["properties"] = {}
        body_insert["properties"]["some_other_wrong_field"] = "test"
        mock_hook.return_value.insert_instance_template.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            body=body_insert,
            request_id=None,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_copy_template_with_updated_nested_fields_should_execute_successfully(self, mock_hook):
        get_template_obj_mock = mock.MagicMock()
        get_template_obj_mock.__class__ = InstanceTemplate
        mock_hook.return_value.get_instance_template.side_effect = [
            NotFound("Error message"),
            get_template_obj_mock,
            get_template_obj_mock,
        ]
        op = ComputeEngineCopyInstanceTemplateOperator(
            project_id=GCP_PROJECT_ID,
            resource_id=GCE_INSTANCE_TEMPLATE_NAME,
            task_id=TASK_ID,
            body_patch={
                "name": GCE_INSTANCE_TEMPLATE_NEW_NAME,
                "properties": {
                    "machine_type": "n1-standard-2",
                },
            },
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            api_version="v1",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        body_insert = deepcopy(GCE_INSTANCE_TEMPLATE_BODY_INSERT)
        body_insert["properties"] = {}
        body_insert["properties"]["machine_type"] = "n1-standard-2"
        mock_hook.return_value.insert_instance_template.assert_called_once_with(
            project_id=GCP_PROJECT_ID, body=body_insert, request_id=None
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_copy_template_with_bigger_array_fields_should_execute_successfully(self, mock_hook):
        get_template_obj_mock = mock.MagicMock()
        get_template_obj_mock.__class__ = InstanceTemplate
        mock_hook.return_value.get_instance_template.side_effect = [
            NotFound("Error message"),
            get_template_obj_mock,
            get_template_obj_mock,
        ]
        op = ComputeEngineCopyInstanceTemplateOperator(
            project_id=GCP_PROJECT_ID,
            resource_id=GCE_INSTANCE_TEMPLATE_NAME,
            task_id=TASK_ID,
            body_patch={
                "name": GCE_INSTANCE_TEMPLATE_NEW_NAME,
                "properties": {
                    "disks": [
                        {
                            "kind": "compute#attachedDisk",
                            "type": "SCRATCH",
                            "licenses": [
                                "Updated String",
                            ],
                        },
                        {
                            "kind": "compute#attachedDisk",
                            "type": "PERSISTENT",
                            "licenses": [
                                "Another String",
                            ],
                        },
                    ],
                },
            },
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        op.execute(context=context)
        mock_hook.assert_called_once_with(
            api_version="v1",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        body_insert = deepcopy(GCE_INSTANCE_TEMPLATE_BODY_INSERT)
        body_insert["properties"] = {}
        body_insert["properties"]["disks"] = [
            {
                "kind": "compute#attachedDisk",
                "type": "SCRATCH",
                "licenses": [
                    "Updated String",
                ],
            },
            {
                "kind": "compute#attachedDisk",
                "type": "PERSISTENT",
                "licenses": [
                    "Another String",
                ],
            },
        ]
        mock_hook.return_value.insert_instance_template.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            body=body_insert,
            request_id=None,
        )

    def test_copy_template_with_missing_name_should_execute_successfully(self):
        with pytest.raises(
            AirflowException,
            match=r"should contain at least name for the new operator in the 'name' field",
        ):
            ComputeEngineCopyInstanceTemplateOperator(
                project_id=GCP_PROJECT_ID,
                resource_id=GCE_INSTANCE_TEMPLATE_NAME,
                request_id=GCE_INSTANCE_TEMPLATE_REQUEST_ID,
                task_id=TASK_ID,
                body_patch={"description": "New description"},
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )


GCE_INSTANCE_GROUP_MANAGER_NAME = "instance-group-test"
GCE_INSTANCE_TEMPLATE_SOURCE_URL = (
    "https://www.googleapis.com/compute/beta/projects/project/global/instanceTemplates/instance-template-test"
)

GCE_INSTANCE_TEMPLATE_OTHER_URL = (
    "https://www.googleapis.com/compute/beta/projects/project"
    "/global/instanceTemplates/instance-template-other"
)

GCE_INSTANCE_TEMPLATE_NON_EXISTING_URL = (
    "https://www.googleapis.com/compute/beta/projects/project"
    "/global/instanceTemplates/instance-template-non-existing"
)

GCE_INSTANCE_TEMPLATE_DESTINATION_URL = (
    "https://www.googleapis.com/compute/beta/projects/project/global/instanceTemplates/instance-template-new"
)

GCE_INSTANCE_GROUP_MANAGER_GET = {
    "kind": "compute#instanceGroupManager",
    "id": "2822359583810032488",
    "creationTimestamp": "2018-10-17T05:39:35.793-07:00",
    "name": GCE_INSTANCE_GROUP_MANAGER_NAME,
    "zone": "https://www.googleapis.com/compute/beta/projects/project/zones/zone",
    "instanceTemplate": GCE_INSTANCE_TEMPLATE_SOURCE_URL,
    "versions": [
        {"name": "v1", "instanceTemplate": GCE_INSTANCE_TEMPLATE_SOURCE_URL, "targetSize": {"calculated": 1}},
        {
            "name": "v2",
            "instanceTemplate": GCE_INSTANCE_TEMPLATE_OTHER_URL,
        },
    ],
    "instanceGroup": GCE_INSTANCE_TEMPLATE_SOURCE_URL,
    "baseInstanceName": GCE_INSTANCE_GROUP_MANAGER_NAME,
    "fingerprint": "BKWB_igCNbQ=",
    "currentActions": {
        "none": 1,
        "creating": 0,
        "creatingWithoutRetries": 0,
        "verifying": 0,
        "recreating": 0,
        "deleting": 0,
        "abandoning": 0,
        "restarting": 0,
        "refreshing": 0,
    },
    "pendingActions": {"creating": 0, "deleting": 0, "recreating": 0, "restarting": 0},
    "targetSize": 1,
    "selfLink": "https://www.googleapis.com/compute/beta/projects/project/zones/"
    "zone/instanceGroupManagers/" + GCE_INSTANCE_GROUP_MANAGER_NAME,
    "autoHealingPolicies": [{"initialDelaySec": 300}],
    "serviceAccount": "198907790164@cloudservices.gserviceaccount.com",
}

GCE_INSTANCE_GROUP_MANAGER_UPDATE_POLICY = {
    "type": "OPPORTUNISTIC",
    "minimalAction": "RESTART",
    "maxSurge": {"fixed": 1},
    "maxUnavailable": {"percent": 10},
    "minReadySec": 1800,
}

GCE_INSTANCE_GROUP_MANAGER_EXPECTED_PATCH = {
    "instanceTemplate": GCE_INSTANCE_TEMPLATE_DESTINATION_URL,
    "versions": [
        {
            "name": "v1",
            "instanceTemplate": GCE_INSTANCE_TEMPLATE_DESTINATION_URL,
            "targetSize": {"calculated": 1},
        },
        {
            "name": "v2",
            "instanceTemplate": GCE_INSTANCE_TEMPLATE_OTHER_URL,
        },
    ],
    "updatePolicy": GCE_INSTANCE_GROUP_MANAGER_UPDATE_POLICY,
}

GCE_INSTANCE_GROUP_MANAGER_REQUEST_ID = "e12d5b48-4826-4ba9-ada6-0cff1e0b36a6"

GCE_INSTANCE_GROUP_MANAGER_BODY_API_CALL = {
    "name": GCE_INSTANCE_GROUP_MANAGER_NAME,
    "base_instance_name": GCE_INSTANCE_GROUP_MANAGER_NAME,
    "instance_template": GCE_INSTANCE_TEMPLATE_SOURCE_URL,
    "target_size": 1,
}


class TestGceInstanceGroupManagerInsert:
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_insert_igm_should_execute_successfully(self, mock_hook):
        get_instance_group_manager_obj_mock = mock.MagicMock()
        get_instance_group_manager_obj_mock.__class__ = InstanceGroupManager
        mock_hook.return_value.get_instance_group_manager.side_effect = [
            NotFound("Error message"),
            get_instance_group_manager_obj_mock,
        ]
        op = ComputeEngineInsertInstanceGroupManagerOperator(
            project_id=GCP_PROJECT_ID,
            body=GCE_INSTANCE_GROUP_MANAGER_BODY_API_CALL,
            zone=GCE_ZONE,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance_group_manager.assert_called_once_with(
            zone=GCE_ZONE,
            project_id=GCP_PROJECT_ID,
            body=GCE_INSTANCE_GROUP_MANAGER_BODY_API_CALL,
            request_id=None,
        )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_insert_igm_should_throw_ex_when_missing_project_id(self, mock_hook):
        with pytest.raises(AirflowException, match=r"The required parameter 'project_id' is missing"):
            ComputeEngineInsertInstanceGroupManagerOperator(
                project_id="",
                body=GCE_INSTANCE_GROUP_MANAGER_BODY_API_CALL,
                zone=GCE_ZONE,
                task_id=TASK_ID,
                retry=RETRY,
                timeout=TIMEOUT,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_insert_igm_should_not_throw_ex_when_project_id_none(self, mock_hook):
        get_instance_group_manager_obj_mock = mock.MagicMock()
        get_instance_group_manager_obj_mock.__class__ = InstanceGroupManager
        mock_hook.return_value.get_instance_group_manager.side_effect = [
            NotFound("Error message"),
            get_instance_group_manager_obj_mock,
        ]
        op = ComputeEngineInsertInstanceGroupManagerOperator(
            body=GCE_INSTANCE_GROUP_MANAGER_BODY_API_CALL,
            zone=GCE_ZONE,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance_group_manager.assert_called_once_with(
            zone=GCE_ZONE,
            project_id=None,
            body=GCE_INSTANCE_GROUP_MANAGER_BODY_API_CALL,
            request_id=None,
        )

    def test_insert_igm_should_throw_ex_when_missing_body(self):
        with pytest.raises((TypeError, AirflowException), match=r"missing keyword argument 'body'"):
            ComputeEngineInsertInstanceGroupManagerOperator(
                zone=GCE_ZONE,
                task_id=TASK_ID,
                retry=RETRY,
                timeout=TIMEOUT,
                metadata=METADATA,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_insert_igm_should_not_throw_ex_when_name_is_templated(self, mock_hook):
        get_instance_group_manager_obj_mock = mock.MagicMock()
        get_instance_group_manager_obj_mock.__class__ = InstanceGroupManager
        mock_hook.return_value.get_instance_group_manager.side_effect = [
            NotFound("Error message"),
            get_instance_group_manager_obj_mock,
        ]
        body_with_templated_name = deepcopy(GCE_INSTANCE_GROUP_MANAGER_BODY_API_CALL)
        body_with_templated_name["name"] = "{{ execution_date }}"
        op = ComputeEngineInsertInstanceGroupManagerOperator(
            project_id=GCP_PROJECT_ID,
            body=body_with_templated_name,
            zone=GCE_ZONE,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.insert_instance_group_manager.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            body=body_with_templated_name,
            request_id=None,
        )


class TestGceInstanceGroupManagerDelete:
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_delete_igm_should_execute_successfully(self, mock_hook):
        op = ComputeEngineDeleteInstanceGroupManagerOperator(
            project_id=GCP_PROJECT_ID,
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            zone=GCE_ZONE,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_instance_group_manager.assert_called_once_with(
            zone=GCE_ZONE,
            request_id=None,
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            project_id=GCP_PROJECT_ID,
        )

    def test_delete_igm_should_throw_ex_when_missing_project_id(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'project_id' is missing"):
            ComputeEngineDeleteInstanceGroupManagerOperator(
                project_id="",
                resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
                zone=GCE_ZONE,
                task_id=TASK_ID,
                retry=RETRY,
                timeout=TIMEOUT,
                metadata=METADATA,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_delete_igm_should_not_throw_ex_when_project_id_none(self, mock_hook):
        op = ComputeEngineDeleteInstanceGroupManagerOperator(
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            zone=GCE_ZONE,
            task_id=TASK_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.delete_instance_group_manager.assert_called_once_with(
            zone=GCE_ZONE,
            request_id=None,
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            project_id=None,
        )

    def test_delete_igm_should_throw_ex_when_missing_resource_id(self):
        with pytest.raises(AirflowException, match=r"The required parameter 'resource_id' is missing"):
            ComputeEngineDeleteInstanceGroupManagerOperator(
                resource_id="",
                zone=GCE_ZONE,
                task_id=TASK_ID,
                retry=RETRY,
                timeout=TIMEOUT,
                metadata=METADATA,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
            )


class TestGceInstanceGroupManagerUpdate:
    @mock.patch(IGM_PATH)
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_update_igm_should_execute_successfully(self, mock_hook, igm):
        get_instance_group_manager_obj_mock = mock.MagicMock()
        get_instance_group_manager_obj_mock.__class__ = InstanceGroupManager
        mock_hook.return_value.get_instance_group_manager.side_effect = [
            get_instance_group_manager_obj_mock,
        ]
        igm.to_dict.return_value = GCE_INSTANCE_GROUP_MANAGER_GET

        op = ComputeEngineInstanceGroupUpdateManagerTemplateOperator(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            task_id=TASK_ID,
            source_template=GCE_INSTANCE_TEMPLATE_SOURCE_URL,
            destination_template=GCE_INSTANCE_TEMPLATE_DESTINATION_URL,
            update_policy=GCE_INSTANCE_GROUP_MANAGER_UPDATE_POLICY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="beta",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.patch_instance_group_manager.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            body=GCE_INSTANCE_GROUP_MANAGER_EXPECTED_PATCH,
            request_id=None,
        )
        assert result

    @mock.patch(IGM_PATH)
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_update_instance_group_missing_project_id_should_execute_successfully(self, mock_hook, igm):
        get_instance_group_manager_obj_mock = mock.MagicMock()
        get_instance_group_manager_obj_mock.__class__ = InstanceGroupManager
        mock_hook.return_value.get_instance_group_manager.side_effect = [
            get_instance_group_manager_obj_mock,
        ]
        igm.to_dict.return_value = GCE_INSTANCE_GROUP_MANAGER_GET
        op = ComputeEngineInstanceGroupUpdateManagerTemplateOperator(
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            task_id=TASK_ID,
            source_template=GCE_INSTANCE_TEMPLATE_SOURCE_URL,
            destination_template=GCE_INSTANCE_TEMPLATE_DESTINATION_URL,
            update_policy=GCE_INSTANCE_GROUP_MANAGER_UPDATE_POLICY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="beta",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.patch_instance_group_manager.assert_called_once_with(
            project_id=None,
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            body=GCE_INSTANCE_GROUP_MANAGER_EXPECTED_PATCH,
            request_id=None,
        )
        assert result

    @mock.patch(IGM_PATH)
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_update_instance_group_no_instance_template_field_should_execute_successfully(
        self, mock_hook, igm
    ):
        instance_group_manager_no_template = deepcopy(GCE_INSTANCE_GROUP_MANAGER_GET)
        del instance_group_manager_no_template["instanceTemplate"]
        get_instance_group_manager_obj_mock = mock.MagicMock()
        get_instance_group_manager_obj_mock.__class__ = InstanceGroupManager
        mock_hook.return_value.get_instance_group_manager.side_effect = [
            get_instance_group_manager_obj_mock,
        ]
        igm.to_dict.return_value = instance_group_manager_no_template
        op = ComputeEngineInstanceGroupUpdateManagerTemplateOperator(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            task_id=TASK_ID,
            source_template=GCE_INSTANCE_TEMPLATE_SOURCE_URL,
            destination_template=GCE_INSTANCE_TEMPLATE_DESTINATION_URL,
            update_policy=GCE_INSTANCE_GROUP_MANAGER_UPDATE_POLICY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="beta",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        expected_patch_no_instance_template = deepcopy(GCE_INSTANCE_GROUP_MANAGER_EXPECTED_PATCH)
        del expected_patch_no_instance_template["instanceTemplate"]
        mock_hook.return_value.patch_instance_group_manager.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            body=expected_patch_no_instance_template,
            request_id=None,
        )
        assert result

    @mock.patch(IGM_PATH)
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_update_instance_group_no_versions_field_should_execute_successfully(self, mock_hook, igm):
        instance_group_manager_no_versions = deepcopy(GCE_INSTANCE_GROUP_MANAGER_GET)
        del instance_group_manager_no_versions["versions"]
        get_instance_group_manager_obj_mock = mock.MagicMock()
        get_instance_group_manager_obj_mock.__class__ = InstanceGroupManager
        mock_hook.return_value.get_instance_group_manager.side_effect = [
            get_instance_group_manager_obj_mock,
        ]
        igm.to_dict.return_value = instance_group_manager_no_versions
        op = ComputeEngineInstanceGroupUpdateManagerTemplateOperator(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            task_id=TASK_ID,
            source_template=GCE_INSTANCE_TEMPLATE_SOURCE_URL,
            destination_template=GCE_INSTANCE_TEMPLATE_DESTINATION_URL,
            update_policy=GCE_INSTANCE_GROUP_MANAGER_UPDATE_POLICY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="beta",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        expected_patch_no_versions = deepcopy(GCE_INSTANCE_GROUP_MANAGER_EXPECTED_PATCH)
        del expected_patch_no_versions["versions"]
        mock_hook.return_value.patch_instance_group_manager.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            body=expected_patch_no_versions,
            request_id=None,
        )
        assert result

    @mock.patch(IGM_PATH)
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_update_instance_group_with_request_id_should_execute_successfully(self, mock_hook, igm):
        get_instance_group_manager_obj_mock = mock.MagicMock()
        get_instance_group_manager_obj_mock.__class__ = InstanceGroupManager
        mock_hook.return_value.get_instance_group_manager.side_effect = [
            get_instance_group_manager_obj_mock,
        ]
        igm.to_dict.return_value = GCE_INSTANCE_GROUP_MANAGER_GET
        op = ComputeEngineInstanceGroupUpdateManagerTemplateOperator(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            task_id=TASK_ID,
            source_template=GCE_INSTANCE_TEMPLATE_SOURCE_URL,
            request_id=GCE_INSTANCE_GROUP_MANAGER_REQUEST_ID,
            destination_template=GCE_INSTANCE_TEMPLATE_DESTINATION_URL,
            update_policy=GCE_INSTANCE_GROUP_MANAGER_UPDATE_POLICY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="beta",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.patch_instance_group_manager.assert_called_once_with(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            body=GCE_INSTANCE_GROUP_MANAGER_EXPECTED_PATCH,
            request_id=GCE_INSTANCE_GROUP_MANAGER_REQUEST_ID,
        )
        assert result

    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_update_instance_group_try_to_use_api_v1_should_throw_ex(self, _):
        with pytest.raises(AirflowException, match=r"Use beta api version or above"):
            ComputeEngineInstanceGroupUpdateManagerTemplateOperator(
                project_id=GCP_PROJECT_ID,
                zone=GCE_ZONE,
                resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
                task_id=TASK_ID,
                api_version="v1",
                source_template=GCE_INSTANCE_TEMPLATE_SOURCE_URL,
                destination_template=GCE_INSTANCE_TEMPLATE_DESTINATION_URL,
            )

    @mock.patch(IGM_PATH)
    @mock.patch(COMPUTE_ENGINE_HOOK_PATH)
    def test_update_instance_group_try_to_use_non_existing_template_should_not_call_patch(
        self, mock_hook, igm
    ):
        get_instance_group_manager_obj_mock = mock.MagicMock()
        get_instance_group_manager_obj_mock.__class__ = InstanceGroupManager
        mock_hook.return_value.get_instance_group_manager.side_effect = [
            get_instance_group_manager_obj_mock,
        ]
        igm.to_dict.return_value = GCE_INSTANCE_GROUP_MANAGER_GET
        op = ComputeEngineInstanceGroupUpdateManagerTemplateOperator(
            project_id=GCP_PROJECT_ID,
            zone=GCE_ZONE,
            resource_id=GCE_INSTANCE_GROUP_MANAGER_NAME,
            task_id=TASK_ID,
            source_template=GCE_INSTANCE_TEMPLATE_NON_EXISTING_URL,
            destination_template=GCE_INSTANCE_TEMPLATE_DESTINATION_URL,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        result = op.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            api_version="beta",
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.patch_instance_group_manager.assert_not_called()
        assert result



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Dataproc Job sensor."""

from __future__ import annotations

import time
from collections.abc import Sequence
from typing import TYPE_CHECKING

from google.api_core.exceptions import ServerError
from google.cloud.dataproc_v1.types import Batch, JobStatus

from airflow.providers.common.compat.sdk import AirflowException, BaseSensorOperator
from airflow.providers.google.cloud.hooks.dataproc import DataprocHook
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class DataprocJobSensor(BaseSensorOperator):
    """
    Check for the state of a previously submitted Dataproc job.

    :param dataproc_job_id: The Dataproc job ID to poll. (templated)
    :param region: Required. The Cloud Dataproc region in which to handle the request. (templated)
    :param project_id: The ID of the google cloud project in which
        to create the cluster. (templated)
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud Platform.
    :param wait_timeout: How many seconds wait for job to be ready.
    """

    template_fields: Sequence[str] = ("project_id", "region", "dataproc_job_id")
    ui_color = "#f0eee4"

    def __init__(
        self,
        *,
        dataproc_job_id: str,
        region: str,
        project_id: str = PROVIDE_PROJECT_ID,
        gcp_conn_id: str = "google_cloud_default",
        wait_timeout: int | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.gcp_conn_id = gcp_conn_id
        self.dataproc_job_id = dataproc_job_id
        self.region = region
        self.wait_timeout = wait_timeout
        self.start_sensor_time: float | None = None

    def execute(self, context: Context) -> None:
        self.start_sensor_time = time.monotonic()
        super().execute(context)

    def _duration(self):
        return time.monotonic() - self.start_sensor_time

    def poke(self, context: Context) -> bool:
        hook = DataprocHook(gcp_conn_id=self.gcp_conn_id)
        if self.wait_timeout:
            try:
                job = hook.get_job(
                    job_id=self.dataproc_job_id, region=self.region, project_id=self.project_id
                )
            except ServerError as err:
                duration = self._duration()
                self.log.info("DURATION RUN: %f", duration)
                if duration > self.wait_timeout:
                    message = (
                        f"Timeout: dataproc job {self.dataproc_job_id} "
                        f"is not ready after {self.wait_timeout}s"
                    )
                    raise AirflowException(message)
                self.log.info("Retrying. Dataproc API returned server error when waiting for job: %s", err)
                return False
        else:
            job = hook.get_job(job_id=self.dataproc_job_id, region=self.region, project_id=self.project_id)

        state = job.status.state
        if state == JobStatus.State.ERROR:
            message = f"Job failed:\n{job}"
            raise AirflowException(message)
        if state in {
            JobStatus.State.CANCELLED,
            JobStatus.State.CANCEL_PENDING,
            JobStatus.State.CANCEL_STARTED,
        }:
            message = f"Job was cancelled:\n{job}"
            raise AirflowException(message)
        if state == JobStatus.State.DONE:
            self.log.debug("Job %s completed successfully.", self.dataproc_job_id)
            return True
        if state == JobStatus.State.ATTEMPT_FAILURE:
            self.log.debug("Job %s attempt has failed.", self.dataproc_job_id)

        self.log.info("Waiting for job %s to complete.", self.dataproc_job_id)
        return False


class DataprocBatchSensor(BaseSensorOperator):
    """
    Check for the state of batch.

    :param batch_id: The Dataproc batch ID to poll. (templated)
    :param region: Required. The Cloud Dataproc region in which to handle the request. (templated)
    :param project_id: The ID of the google cloud project in which
        to create the cluster. (templated)
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud Platform.
    :param wait_timeout: How many seconds wait for job to be ready.
    """

    template_fields: Sequence[str] = ("project_id", "region", "batch_id")
    ui_color = "#f0eee4"

    def __init__(
        self,
        *,
        batch_id: str,
        region: str,
        project_id: str = PROVIDE_PROJECT_ID,
        gcp_conn_id: str = "google_cloud_default",
        wait_timeout: int | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.batch_id = batch_id
        self.project_id = project_id
        self.gcp_conn_id = gcp_conn_id
        self.region = region
        self.wait_timeout = wait_timeout
        self.start_sensor_time: float | None = None

    def execute(self, context: Context) -> None:
        self.start_sensor_time = time.monotonic()
        super().execute(context)

    def _duration(self):
        return time.monotonic() - self.start_sensor_time

    def poke(self, context: Context) -> bool:
        hook = DataprocHook(gcp_conn_id=self.gcp_conn_id)
        if self.wait_timeout:
            try:
                batch = hook.get_batch(batch_id=self.batch_id, region=self.region, project_id=self.project_id)
            except ServerError as err:
                duration = self._duration()
                self.log.info("DURATION RUN: %f", duration)

                if duration > self.wait_timeout:
                    raise AirflowException(
                        f"Timeout: dataproc batch {self.batch_id} is not ready after {self.wait_timeout}s"
                    )
                self.log.info("Retrying. Dataproc API returned server error when waiting for batch: %s", err)
                return False
        else:
            batch = hook.get_batch(batch_id=self.batch_id, region=self.region, project_id=self.project_id)

        state = batch.state
        if state == Batch.State.FAILED:
            message = "Batch failed"
            raise AirflowException(message)
        if state in {
            Batch.State.CANCELLED,
            Batch.State.CANCELLING,
        }:
            message = "Batch was cancelled."
            raise AirflowException(message)
        if state == Batch.State.SUCCEEDED:
            self.log.debug("Batch %s completed successfully.", self.batch_id)
            return True

        self.log.info("Waiting for the batch %s to complete.", self.batch_id)
        return False



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Objects relating to sourcing connections from Google Cloud Secrets Manager."""

from __future__ import annotations

import logging
from collections.abc import Sequence

from google.auth.exceptions import DefaultCredentialsError

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.cloud._internal_client.secret_manager_client import _SecretManagerClient
from airflow.providers.google.cloud.utils.credentials_provider import (
    _get_target_principal_and_delegates,
    get_credentials_and_project_id,
)
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID
from airflow.secrets import BaseSecretsBackend
from airflow.utils.log.logging_mixin import LoggingMixin

log = logging.getLogger(__name__)

SECRET_ID_PATTERN = r"^[a-zA-Z0-9-_]*$"


class CloudSecretManagerBackend(BaseSecretsBackend, LoggingMixin):
    """
    Retrieves Connection object from Google Cloud Secrets Manager.

    Configurable via ``airflow.cfg`` as follows:

    .. code-block:: ini

        [secrets]
        backend = airflow.providers.google.cloud.secrets.secret_manager.CloudSecretManagerBackend
        backend_kwargs = {"connections_prefix": "airflow-connections", "sep": "-"}

    For example, if the Secrets Manager secret id is ``airflow-connections-smtp_default``, this would be
    accessible if you provide ``{"connections_prefix": "airflow-connections", "sep": "-"}`` and request
    conn_id ``smtp_default``.

    If the Secrets Manager secret id is ``airflow-variables-hello``, this would be
    accessible if you provide ``{"variables_prefix": "airflow-variables", "sep": "-"}`` and request
    Variable Key ``hello``.

    The full secret id should follow the pattern "[a-zA-Z0-9-_]".

    :param connections_prefix: Specifies the prefix of the secret to read to get Connections.
        If set to None (null), requests for connections will not be sent to GCP Secrets Manager
    :param variables_prefix: Specifies the prefix of the secret to read to get Variables.
        If set to None (null), requests for variables will not be sent to GCP Secrets Manager
    :param config_prefix: Specifies the prefix of the secret to read to get Airflow Configurations
        containing secrets.
        If set to None (null), requests for configurations will not be sent to GCP Secrets Manager
    :param gcp_key_path: Path to Google Cloud Service Account key file (JSON). Mutually exclusive with
        gcp_keyfile_dict. use default credentials in the current environment if not provided.
    :param gcp_keyfile_dict: Dictionary of keyfile parameters. Mutually exclusive with gcp_key_path.
    :param gcp_credential_config_file: File path to or content of a GCP credential configuration file.
    :param gcp_scopes: Comma-separated string containing OAuth2 scopes
    :param project_id: Project ID to read the secrets from. If not passed, the project ID from credentials
        will be used.
    :param sep: Separator used to concatenate connections_prefix and conn_id. Default: "-"
    :param impersonation_chain: Optional service account to impersonate using
        short-term credentials, or chained list of accounts required to get the
        access token of the last account in the list, which will be impersonated
        in the request. If set as a string, the account must grant the
        originating account the Service Account Token Creator IAM role. If set
        as a sequence, the identities from the list must grant Service Account
        Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account.
    """

    def __init__(
        self,
        connections_prefix: str = "airflow-connections",
        variables_prefix: str = "airflow-variables",
        config_prefix: str = "airflow-config",
        gcp_keyfile_dict: dict | None = None,
        gcp_key_path: str | None = None,
        gcp_credential_config_file: dict[str, str] | str | None = None,
        gcp_scopes: str | None = None,
        project_id: str = PROVIDE_PROJECT_ID,
        sep: str = "-",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.connections_prefix = connections_prefix
        self.variables_prefix = variables_prefix
        self.config_prefix = config_prefix
        self.sep = sep
        if connections_prefix is not None:
            if not self._is_valid_prefix_and_sep():
                raise AirflowException(
                    "`connections_prefix`, `variables_prefix` and `sep` should "
                    f"follows that pattern {SECRET_ID_PATTERN}"
                )
        try:
            if impersonation_chain:
                target_principal, delegates = _get_target_principal_and_delegates(impersonation_chain)
            else:
                target_principal = None
                delegates = None

            self.credentials, self.project_id = get_credentials_and_project_id(
                keyfile_dict=gcp_keyfile_dict,
                key_path=gcp_key_path,
                credential_config_file=gcp_credential_config_file,
                scopes=gcp_scopes,
                target_principal=target_principal,
                delegates=delegates,
            )
        except (DefaultCredentialsError, FileNotFoundError):
            log.exception(
                "Unable to load credentials for GCP Secret Manager. "
                "Make sure that the keyfile path or dictionary, credential configuration file, "
                "or GOOGLE_APPLICATION_CREDENTIALS environment variable is correct and properly configured."
            )

        # In case project id provided
        if project_id:
            self.project_id = project_id

    @property
    def client(self) -> _SecretManagerClient:
        """
        Property returning secret client.

        :return: Secrets client
        """
        return _SecretManagerClient(credentials=self.credentials)

    def _is_valid_prefix_and_sep(self) -> bool:
        prefix = self.connections_prefix + self.sep
        return _SecretManagerClient.is_valid_secret_name(prefix)

    def get_conn_value(self, conn_id: str, team_name: str | None = None) -> str | None:
        """
        Get serialized representation of Connection.

        :param conn_id: connection id
        :param team_name: Team name associated to the task trying to access the connection (if any)
        """
        if self.connections_prefix is None:
            return None

        return self._get_secret(self.connections_prefix, conn_id)

    def get_variable(self, key: str, team_name: str | None = None) -> str | None:
        """
        Get Airflow Variable from Environment Variable.

        :param key: Variable Key
        :param team_name: Team name associated to the task trying to access the variable (if any)
        :return: Variable Value
        """
        if self.variables_prefix is None:
            return None

        return self._get_secret(self.variables_prefix, key)

    def get_config(self, key: str) -> str | None:
        """
        Get Airflow Configuration.

        :param key: Configuration Option Key
        :return: Configuration Option Value
        """
        if self.config_prefix is None:
            return None

        return self._get_secret(self.config_prefix, key)

    def _get_secret(self, path_prefix: str, secret_id: str) -> str | None:
        """
        Get secret value from the SecretManager based on prefix.

        :param path_prefix: Prefix for the Path to get Secret
        :param secret_id: Secret Key
        """
        secret_id = self.build_path(path_prefix, secret_id, self.sep)
        return self.client.get_secret(secret_id=secret_id, project_id=self.project_id)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import datetime
import json
import re
from unittest import mock

import pendulum
from google.protobuf.timestamp_pb2 import Timestamp

from airflow.models.dagrun import DagRun
from airflow.providers.common.compat.sdk import Context
from airflow.providers.google.cloud.operators.workflows import (
    WorkflowsCancelExecutionOperator,
    WorkflowsCreateExecutionOperator,
    WorkflowsCreateWorkflowOperator,
    WorkflowsDeleteWorkflowOperator,
    WorkflowsGetExecutionOperator,
    WorkflowsGetWorkflowOperator,
    WorkflowsListExecutionsOperator,
    WorkflowsListWorkflowsOperator,
    WorkflowsUpdateWorkflowOperator,
)
from airflow.utils.hashlib_wrapper import md5

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

BASE_PATH = "airflow.providers.google.cloud.operators.workflows.{}"
LOCATION = "europe-west1"
WORKFLOW_ID = "workflow_id"
EXECUTION_ID = "execution_id"
WORKFLOW = {"aa": "bb"}
EXECUTION = {"ccc": "ddd"}
PROJECT_ID = "airflow-testing"
METADATA = None
TIMEOUT = None
RETRY = None
FILTER_ = "aaaa"
ORDER_BY = "bbb"
UPDATE_MASK = "aaa,bbb"
GCP_CONN_ID = "test-conn"
IMPERSONATION_CHAIN = None


class TestWorkflowsCreateWorkflowOperator:
    @mock.patch(BASE_PATH.format("Workflow"))
    @mock.patch(BASE_PATH.format("WorkflowsHook"))
    def test_execute(self, mock_hook, mock_object):
        op = WorkflowsCreateWorkflowOperator(
            task_id="test_task",
            workflow=WORKFLOW,
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        result = op.execute(context=context)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.create_workflow.assert_called_once_with(
            workflow=WORKFLOW,
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert result == mock_object.to_dict.return_value

    def test_execute_without_workflow_id(self):
        op = WorkflowsCreateWorkflowOperator(
            task_id="test_task",
            workflow=WORKFLOW,
            workflow_id="",
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        hash_base = json.dumps(WORKFLOW, sort_keys=True)
        date = pendulum.datetime(2025, 1, 1)
        ctx = Context(logical_date=date)
        if AIRFLOW_V_3_0_PLUS:
            ctx["dag_run"] = DagRun(run_after=date)
        expected = md5(f"airflow_{op.dag_id}_test_task_{date.isoformat()}_{hash_base}".encode()).hexdigest()
        assert op._workflow_id(ctx) == re.sub(r"[:\-+.]", "_", expected)

        if AIRFLOW_V_3_0_PLUS:
            ctx = Context(dag_run=DagRun(run_after=date))
            assert op._workflow_id(ctx) == re.sub(r"[:\-+.]", "_", expected)


class TestWorkflowsUpdateWorkflowOperator:
    @mock.patch(BASE_PATH.format("Workflow"))
    @mock.patch(BASE_PATH.format("WorkflowsHook"))
    def test_execute(self, mock_hook, mock_object):
        op = WorkflowsUpdateWorkflowOperator(
            task_id="test_task",
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            update_mask=UPDATE_MASK,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        result = op.execute(context=context)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.get_workflow.assert_called_once_with(
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        mock_hook.return_value.update_workflow.assert_called_once_with(
            workflow=mock_hook.return_value.get_workflow.return_value,
            update_mask=UPDATE_MASK,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert result == mock_object.to_dict.return_value


class TestWorkflowsDeleteWorkflowOperator:
    @mock.patch(BASE_PATH.format("WorkflowsHook"))
    def test_execute(
        self,
        mock_hook,
    ):
        op = WorkflowsDeleteWorkflowOperator(
            task_id="test_task",
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute({})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.delete_workflow.assert_called_once_with(
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )


class TestWorkflowsListWorkflowsOperator:
    @mock.patch(BASE_PATH.format("Workflow"))
    @mock.patch(BASE_PATH.format("WorkflowsHook"))
    def test_execute(self, mock_hook, mock_object):
        timestamp = Timestamp()
        timestamp.FromDatetime(
            datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(minutes=5)
        )
        workflow_mock = mock.MagicMock()
        workflow_mock.start_time = timestamp
        mock_hook.return_value.list_workflows.return_value = [workflow_mock]

        op = WorkflowsListWorkflowsOperator(
            task_id="test_task",
            location=LOCATION,
            project_id=PROJECT_ID,
            filter_=FILTER_,
            order_by=ORDER_BY,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        result = op.execute(context=context)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.list_workflows.assert_called_once_with(
            location=LOCATION,
            project_id=PROJECT_ID,
            filter_=FILTER_,
            order_by=ORDER_BY,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert result == [mock_object.to_dict.return_value]


class TestWorkflowsGetWorkflowOperator:
    @mock.patch(BASE_PATH.format("Workflow"))
    @mock.patch(BASE_PATH.format("WorkflowsHook"))
    def test_execute(self, mock_hook, mock_object):
        op = WorkflowsGetWorkflowOperator(
            task_id="test_task",
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        result = op.execute(context=context)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.get_workflow.assert_called_once_with(
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert result == mock_object.to_dict.return_value


class TestWorkflowExecutionsCreateExecutionOperator:
    @mock.patch(BASE_PATH.format("Execution"))
    @mock.patch(BASE_PATH.format("WorkflowsHook"))
    @mock.patch(BASE_PATH.format("WorkflowsExecutionLink.persist"))
    def test_execute(self, mock_link_persist, mock_hook, mock_object):
        mock_hook.return_value.create_execution.return_value.name = "name/execution_id"
        op = WorkflowsCreateExecutionOperator(
            task_id="test_task",
            workflow_id=WORKFLOW_ID,
            execution=EXECUTION,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        result = op.execute(context=context)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.create_execution.assert_called_once_with(
            workflow_id=WORKFLOW_ID,
            execution=EXECUTION,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )
        mock_link_persist.assert_called_with(
            context=context,
            location_id=LOCATION,
            workflow_id=WORKFLOW_ID,
            execution_id=EXECUTION_ID,
            project_id=PROJECT_ID,
        )
        assert result == mock_object.to_dict.return_value


class TestWorkflowExecutionsCancelExecutionOperator:
    @mock.patch(BASE_PATH.format("Execution"))
    @mock.patch(BASE_PATH.format("WorkflowsHook"))
    def test_execute(self, mock_hook, mock_object):
        op = WorkflowsCancelExecutionOperator(
            task_id="test_task",
            workflow_id=WORKFLOW_ID,
            execution_id=EXECUTION_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        result = op.execute(context=context)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.cancel_execution.assert_called_once_with(
            workflow_id=WORKFLOW_ID,
            execution_id=EXECUTION_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert result == mock_object.to_dict.return_value


class TestWorkflowExecutionsListExecutionsOperator:
    @mock.patch(BASE_PATH.format("Execution"))
    @mock.patch(BASE_PATH.format("WorkflowsHook"))
    def test_execute(self, mock_hook, mock_object):
        start_date_filter = datetime.datetime.now(tz=datetime.timezone.utc) + datetime.timedelta(minutes=5)
        execution_mock = mock.MagicMock()
        execution_mock.start_time = start_date_filter
        mock_hook.return_value.list_executions.return_value = [execution_mock]

        op = WorkflowsListExecutionsOperator(
            task_id="test_task",
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        result = op.execute(context=context)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.list_executions.assert_called_once_with(
            workflow_id=WORKFLOW_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert result == [mock_object.to_dict.return_value]


class TestWorkflowExecutionsGetExecutionOperator:
    @mock.patch(BASE_PATH.format("Execution"))
    @mock.patch(BASE_PATH.format("WorkflowsHook"))
    def test_execute(self, mock_hook, mock_object):
        op = WorkflowsGetExecutionOperator(
            task_id="test_task",
            workflow_id=WORKFLOW_ID,
            execution_id=EXECUTION_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        context = mock.MagicMock()
        result = op.execute(context=context)

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.get_execution.assert_called_once_with(
            workflow_id=WORKFLOW_ID,
            execution_id=EXECUTION_ID,
            location=LOCATION,
            project_id=PROJECT_ID,
            retry=RETRY,
            timeout=TIMEOUT,
            metadata=METADATA,
        )

        assert result == mock_object.to_dict.return_value



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Cloud Vertex AI hook."""

from __future__ import annotations

import asyncio
from collections.abc import Sequence
from typing import TYPE_CHECKING

from google.api_core.client_options import ClientOptions
from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.cloud.aiplatform import BatchPredictionJob, Model, explain
from google.cloud.aiplatform_v1 import JobServiceAsyncClient, JobServiceClient, JobState, types

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.common.consts import CLIENT_INFO
from airflow.providers.google.common.hooks.base_google import GoogleBaseAsyncHook, GoogleBaseHook
from airflow.providers.google.common.hooks.operation_helpers import OperationHelper

if TYPE_CHECKING:
    from google.api_core.operation import Operation
    from google.api_core.retry import AsyncRetry, Retry
    from google.cloud.aiplatform_v1.services.job_service.pagers import ListBatchPredictionJobsPager


class BatchPredictionJobHook(GoogleBaseHook, OperationHelper):
    """Hook for Google Cloud Vertex AI Batch Prediction Job APIs."""

    def __init__(
        self,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(
            gcp_conn_id=gcp_conn_id,
            impersonation_chain=impersonation_chain,
            **kwargs,
        )
        self._batch_prediction_job: BatchPredictionJob | None = None

    def get_job_service_client(self, region: str | None = None) -> JobServiceClient:
        """Return JobServiceClient object."""
        if region and region != "global":
            client_options = ClientOptions(api_endpoint=f"{region}-aiplatform.googleapis.com:443")
        else:
            client_options = ClientOptions()

        return JobServiceClient(
            credentials=self.get_credentials(), client_info=CLIENT_INFO, client_options=client_options
        )

    @staticmethod
    def extract_batch_prediction_job_id(obj: dict) -> str:
        """Return unique id of the batch_prediction_job."""
        return obj["name"].rpartition("/")[-1]

    def cancel_batch_prediction_job(self) -> None:
        """Cancel BatchPredictionJob."""
        if self._batch_prediction_job:
            self._batch_prediction_job.cancel()

    @GoogleBaseHook.fallback_to_default_project_id
    def create_batch_prediction_job(
        self,
        project_id: str,
        region: str,
        job_display_name: str,
        model_name: str | Model,
        instances_format: str = "jsonl",
        predictions_format: str = "jsonl",
        gcs_source: str | Sequence[str] | None = None,
        bigquery_source: str | None = None,
        gcs_destination_prefix: str | None = None,
        bigquery_destination_prefix: str | None = None,
        model_parameters: dict | None = None,
        machine_type: str | None = None,
        accelerator_type: str | None = None,
        accelerator_count: int | None = None,
        starting_replica_count: int | None = None,
        max_replica_count: int | None = None,
        generate_explanation: bool | None = False,
        explanation_metadata: explain.ExplanationMetadata | None = None,
        explanation_parameters: explain.ExplanationParameters | None = None,
        labels: dict[str, str] | None = None,
        encryption_spec_key_name: str | None = None,
        sync: bool = True,
        create_request_timeout: float | None = None,
        batch_size: int | None = None,
    ) -> BatchPredictionJob:
        """
        Create a batch prediction job.

        :param project_id: Required. Project to run training in.
        :param region: Required. Location to run training in.
        :param job_display_name: Required. The user-defined name of the BatchPredictionJob. The name can be
            up to 128 characters long and can consist of any UTF-8 characters.
        :param model_name: Required. A fully-qualified model resource name or model ID.
        :param instances_format: Required. The format in which instances are provided. Must be one of the
            formats listed in `Model.supported_input_storage_formats`. Default is "jsonl" when using
            `gcs_source`. If a `bigquery_source` is provided, this is overridden to "bigquery".
        :param predictions_format: Required. The format in which Vertex AI outputs the predictions, must be
            one of the formats specified in `Model.supported_output_storage_formats`. Default is "jsonl" when
            using `gcs_destination_prefix`. If a `bigquery_destination_prefix` is provided, this is
            overridden to "bigquery".
        :param gcs_source: Google Cloud Storage URI(-s) to your instances to run batch prediction on. They
            must match `instances_format`. May contain wildcards. For more information on wildcards, see
            https://cloud.google.com/storage/docs/gsutil/addlhelp/WildcardNames.
        :param bigquery_source: BigQuery URI to a table, up to 2000 characters long.
            For example: `bq://projectId.bqDatasetId.bqTableId`
        :param gcs_destination_prefix: The Google Cloud Storage location of the directory where the output is
            to be written to. In the given directory a new directory is created. Its name is
            ``prediction-<model-display-name>-<job-create-time>``, where timestamp is in
            YYYY-MM-DDThh:mm:ss.sssZ ISO-8601 format. Inside of it files ``predictions_0001.<extension>``,
            ``predictions_0002.<extension>``, ..., ``predictions_N.<extension>`` are created where
            ``<extension>`` depends on chosen ``predictions_format``, and N may equal 0001 and depends on the
            total number of successfully predicted instances. If the Model has both ``instance`` and
            ``prediction`` schemata defined then each such file contains predictions as per the
            ``predictions_format``. If prediction for any instance failed (partially or completely), then an
            additional ``errors_0001.<extension>``, ``errors_0002.<extension>``,..., ``errors_N.<extension>``
            files are created (N depends on total number of failed predictions). These files contain the
            failed instances, as per their schema, followed by an additional ``error`` field which as value
            has ```google.rpc.Status`` <Status>`__ containing only ``code`` and ``message`` fields.
        :param bigquery_destination_prefix: The BigQuery project location where the output is to be written
            to. In the given project a new dataset is created with name
            ``prediction_<model-display-name>_<job-create-time>`` where is made BigQuery-dataset-name
            compatible (for example, most special characters become underscores), and timestamp is in
            YYYY_MM_DDThh_mm_ss_sssZ "based on ISO-8601" format. In the dataset two tables will be created,
            ``predictions``, and ``errors``. If the Model has both ``instance`` and ``prediction`` schemata
            defined then the tables have columns as follows: The ``predictions`` table contains instances for
            which the prediction succeeded, it has columns as per a concatenation of the Model's instance and
            prediction schemata. The ``errors`` table contains rows for which the prediction has failed, it
            has instance columns, as per the instance schema, followed by a single "errors" column, which as
            values has ```google.rpc.Status`` <Status>`__ represented as a STRUCT, and containing only
            ``code`` and ``message``.
        :param model_parameters: The parameters that govern the predictions. The schema of the parameters may
            be specified via the Model's `parameters_schema_uri`.
        :param machine_type: The type of machine for running batch prediction on dedicated resources. Not
            specifying machine type will result in batch prediction job being run with automatic resources.
        :param accelerator_type: The type of accelerator(s) that may be attached to the machine as per
            `accelerator_count`. Only used if `machine_type` is set.
        :param accelerator_count: The number of accelerators to attach to the `machine_type`. Only used if
            `machine_type` is set.
        :param starting_replica_count: The number of machine replicas used at the start of the batch
            operation. If not set, Vertex AI decides starting number, not greater than `max_replica_count`.
            Only used if `machine_type` is set.
        :param max_replica_count: The maximum number of machine replicas the batch operation may be scaled
            to. Only used if `machine_type` is set. Default is 10.
        :param generate_explanation: Optional. Generate explanation along with the batch prediction results.
            This will cause the batch prediction output to include explanations based on the
            `prediction_format`:
            - `bigquery`: output includes a column named `explanation`. The value is a struct that conforms
            to the [aiplatform.gapic.Explanation] object.
            - `jsonl`: The JSON objects on each line include an additional entry keyed `explanation`. The
            value of the entry is a JSON object that conforms to the [aiplatform.gapic.Explanation] object.
            - `csv`: Generating explanations for CSV format is not supported.
        :param explanation_metadata: Optional. Explanation metadata configuration for this
            BatchPredictionJob. Can be specified only if `generate_explanation` is set to `True`.
            This value overrides the value of `Model.explanation_metadata`. All fields of
            `explanation_metadata` are optional in the request. If a field of the `explanation_metadata`
            object is not populated, the corresponding field of the `Model.explanation_metadata` object is
            inherited. For more details, see `Ref docs <http://tinyurl.com/1igh60kt>`
        :param explanation_parameters: Optional. Parameters to configure explaining for Model's predictions.
            Can be specified only if `generate_explanation` is set to `True`.
            This value overrides the value of `Model.explanation_parameters`. All fields of
            `explanation_parameters` are optional in the request. If a field of the `explanation_parameters`
            object is not populated, the corresponding field of the `Model.explanation_parameters` object is
            inherited. For more details, see `Ref docs <http://tinyurl.com/1an4zake>`
        :param labels: Optional. The labels with user-defined metadata to organize your BatchPredictionJobs.
            Label keys and values can be no longer than 64 characters (Unicode codepoints), can only contain
            lowercase letters, numeric characters, underscores and dashes. International characters are
            allowed. See https://goo.gl/xmQnxf for more information and examples of labels.
        :param encryption_spec_key_name: Optional. The Cloud KMS resource identifier of the customer managed
            encryption key used to protect the job. Has the form:
            ``projects/my-project/locations/my-region/keyRings/my-kr/cryptoKeys/my-key``. The key needs to be
            in the same region as where the compute resource is created.
            If this is set, then all resources created by the BatchPredictionJob will be encrypted with the
            provided encryption key.
            Overrides encryption_spec_key_name set in aiplatform.init.
        :param sync: Whether to execute this method synchronously. If False, this method will be executed in
            concurrent Future and any downstream object will be immediately returned and synced when the
            Future has completed.
        :param create_request_timeout: Optional. The timeout for the create request in seconds.
        :param batch_size: Optional. The number of the records (e.g. instances)
            of the operation given in each batch
            to a machine replica. Machine type, and size of a single record should be considered
            when setting this parameter, higher value speeds up the batch operation's execution,
            but too high value will result in a whole batch not fitting in a machine's memory,
            and the whole operation will fail.
            The default value is same as in the aiplatform's BatchPredictionJob.
        """
        self._batch_prediction_job = BatchPredictionJob.create(
            job_display_name=job_display_name,
            model_name=model_name,
            instances_format=instances_format,
            predictions_format=predictions_format,
            gcs_source=gcs_source,
            bigquery_source=bigquery_source,
            gcs_destination_prefix=gcs_destination_prefix,
            bigquery_destination_prefix=bigquery_destination_prefix,
            model_parameters=model_parameters,
            machine_type=machine_type,
            accelerator_type=accelerator_type,
            accelerator_count=accelerator_count,
            starting_replica_count=starting_replica_count,
            max_replica_count=max_replica_count,
            generate_explanation=generate_explanation,
            explanation_metadata=explanation_metadata,
            explanation_parameters=explanation_parameters,
            labels=labels,
            project=project_id,
            location=region,
            credentials=self.get_credentials(),
            encryption_spec_key_name=encryption_spec_key_name,
            sync=sync,
            create_request_timeout=create_request_timeout,
            batch_size=batch_size,
        )
        return self._batch_prediction_job

    @GoogleBaseHook.fallback_to_default_project_id
    def submit_batch_prediction_job(
        self,
        project_id: str,
        region: str,
        job_display_name: str,
        model_name: str | Model,
        instances_format: str = "jsonl",
        predictions_format: str = "jsonl",
        gcs_source: str | Sequence[str] | None = None,
        bigquery_source: str | None = None,
        gcs_destination_prefix: str | None = None,
        bigquery_destination_prefix: str | None = None,
        model_parameters: dict | None = None,
        machine_type: str | None = None,
        accelerator_type: str | None = None,
        accelerator_count: int | None = None,
        starting_replica_count: int | None = None,
        max_replica_count: int | None = None,
        generate_explanation: bool | None = False,
        explanation_metadata: explain.ExplanationMetadata | None = None,
        explanation_parameters: explain.ExplanationParameters | None = None,
        labels: dict[str, str] | None = None,
        encryption_spec_key_name: str | None = None,
        create_request_timeout: float | None = None,
        batch_size: int | None = None,
    ) -> BatchPredictionJob:
        """
        Create a batch prediction job.

        :param project_id: Required. Project to run training in.
        :param region: Required. Location to run training in.
        :param job_display_name: Required. The user-defined name of the BatchPredictionJob. The name can be
            up to 128 characters long and can consist of any UTF-8 characters.
        :param model_name: Required. A fully-qualified model resource name or model ID.
        :param instances_format: Required. The format in which instances are provided. Must be one of the
            formats listed in `Model.supported_input_storage_formats`. Default is "jsonl" when using
            `gcs_source`. If a `bigquery_source` is provided, this is overridden to "bigquery".
        :param predictions_format: Required. The format in which Vertex AI outputs the predictions, must be
            one of the formats specified in `Model.supported_output_storage_formats`. Default is "jsonl" when
            using `gcs_destination_prefix`. If a `bigquery_destination_prefix` is provided, this is
            overridden to "bigquery".
        :param gcs_source: Google Cloud Storage URI(-s) to your instances to run batch prediction on. They
            must match `instances_format`. May contain wildcards. For more information on wildcards, see
            https://cloud.google.com/storage/docs/gsutil/addlhelp/WildcardNames.
        :param bigquery_source: BigQuery URI to a table, up to 2000 characters long.
            For example: `bq://projectId.bqDatasetId.bqTableId`
        :param gcs_destination_prefix: The Google Cloud Storage location of the directory where the output is
            to be written to. In the given directory a new directory is created. Its name is
            ``prediction-<model-display-name>-<job-create-time>``, where timestamp is in
            YYYY-MM-DDThh:mm:ss.sssZ ISO-8601 format. Inside of it files ``predictions_0001.<extension>``,
            ``predictions_0002.<extension>``, ..., ``predictions_N.<extension>`` are created where
            ``<extension>`` depends on chosen ``predictions_format``, and N may equal 0001 and depends on the
            total number of successfully predicted instances. If the Model has both ``instance`` and
            ``prediction`` schemata defined then each such file contains predictions as per the
            ``predictions_format``. If prediction for any instance failed (partially or completely), then an
            additional ``errors_0001.<extension>``, ``errors_0002.<extension>``,..., ``errors_N.<extension>``
            files are created (N depends on total number of failed predictions). These files contain the
            failed instances, as per their schema, followed by an additional ``error`` field which as value
            has ```google.rpc.Status`` <Status>`__ containing only ``code`` and ``message`` fields.
        :param bigquery_destination_prefix: The BigQuery project location where the output is to be written
            to. In the given project a new dataset is created with name
            ``prediction_<model-display-name>_<job-create-time>`` where is made BigQuery-dataset-name
            compatible (for example, most special characters become underscores), and timestamp is in
            YYYY_MM_DDThh_mm_ss_sssZ "based on ISO-8601" format. In the dataset two tables will be created,
            ``predictions``, and ``errors``. If the Model has both ``instance`` and ``prediction`` schemata
            defined then the tables have columns as follows: The ``predictions`` table contains instances for
            which the prediction succeeded, it has columns as per a concatenation of the Model's instance and
            prediction schemata. The ``errors`` table contains rows for which the prediction has failed, it
            has instance columns, as per the instance schema, followed by a single "errors" column, which as
            values has ```google.rpc.Status`` <Status>`__ represented as a STRUCT, and containing only
            ``code`` and ``message``.
        :param model_parameters: The parameters that govern the predictions. The schema of the parameters may
            be specified via the Model's `parameters_schema_uri`.
        :param machine_type: The type of machine for running batch prediction on dedicated resources. Not
            specifying machine type will result in batch prediction job being run with automatic resources.
        :param accelerator_type: The type of accelerator(s) that may be attached to the machine as per
            `accelerator_count`. Only used if `machine_type` is set.
        :param accelerator_count: The number of accelerators to attach to the `machine_type`. Only used if
            `machine_type` is set.
        :param starting_replica_count: The number of machine replicas used at the start of the batch
            operation. If not set, Vertex AI decides starting number, not greater than `max_replica_count`.
            Only used if `machine_type` is set.
        :param max_replica_count: The maximum number of machine replicas the batch operation may be scaled
            to. Only used if `machine_type` is set. Default is 10.
        :param generate_explanation: Optional. Generate explanation along with the batch prediction results.
            This will cause the batch prediction output to include explanations based on the
            `prediction_format`:
            - `bigquery`: output includes a column named `explanation`. The value is a struct that conforms
            to the [aiplatform.gapic.Explanation] object.
            - `jsonl`: The JSON objects on each line include an additional entry keyed `explanation`. The
            value of the entry is a JSON object that conforms to the [aiplatform.gapic.Explanation] object.
            - `csv`: Generating explanations for CSV format is not supported.
        :param explanation_metadata: Optional. Explanation metadata configuration for this
            BatchPredictionJob. Can be specified only if `generate_explanation` is set to `True`.
            This value overrides the value of `Model.explanation_metadata`. All fields of
            `explanation_metadata` are optional in the request. If a field of the `explanation_metadata`
            object is not populated, the corresponding field of the `Model.explanation_metadata` object is
            inherited. For more details, see `Ref docs <http://tinyurl.com/1igh60kt>`
        :param explanation_parameters: Optional. Parameters to configure explaining for Model's predictions.
            Can be specified only if `generate_explanation` is set to `True`.
            This value overrides the value of `Model.explanation_parameters`. All fields of
            `explanation_parameters` are optional in the request. If a field of the `explanation_parameters`
            object is not populated, the corresponding field of the `Model.explanation_parameters` object is
            inherited. For more details, see `Ref docs <http://tinyurl.com/1an4zake>`
        :param labels: Optional. The labels with user-defined metadata to organize your BatchPredictionJobs.
            Label keys and values can be no longer than 64 characters (Unicode codepoints), can only contain
            lowercase letters, numeric characters, underscores and dashes. International characters are
            allowed. See https://goo.gl/xmQnxf for more information and examples of labels.
        :param encryption_spec_key_name: Optional. The Cloud KMS resource identifier of the customer managed
            encryption key used to protect the job. Has the form:
            ``projects/my-project/locations/my-region/keyRings/my-kr/cryptoKeys/my-key``. The key needs to be
            in the same region as where the compute resource is created.
            If this is set, then all resources created by the BatchPredictionJob will be encrypted with the
            provided encryption key.
            Overrides encryption_spec_key_name set in aiplatform.init.
        :param create_request_timeout: Optional. The timeout for the create request in seconds.
        :param batch_size: Optional. The number of the records (e.g. instances)
            of the operation given in each batch
            to a machine replica. Machine type, and size of a single record should be considered
            when setting this parameter, higher value speeds up the batch operation's execution,
            but too high value will result in a whole batch not fitting in a machine's memory,
            and the whole operation will fail.
            The default value is same as in the aiplatform's BatchPredictionJob.
        """
        self._batch_prediction_job = BatchPredictionJob.submit(
            job_display_name=job_display_name,
            model_name=model_name,
            instances_format=instances_format,
            predictions_format=predictions_format,
            gcs_source=gcs_source,
            bigquery_source=bigquery_source,
            gcs_destination_prefix=gcs_destination_prefix,
            bigquery_destination_prefix=bigquery_destination_prefix,
            model_parameters=model_parameters,
            machine_type=machine_type,
            accelerator_type=accelerator_type,
            accelerator_count=accelerator_count,
            starting_replica_count=starting_replica_count,
            max_replica_count=max_replica_count,
            generate_explanation=generate_explanation,
            explanation_metadata=explanation_metadata,
            explanation_parameters=explanation_parameters,
            labels=labels,
            project=project_id,
            location=region,
            credentials=self.get_credentials(),
            encryption_spec_key_name=encryption_spec_key_name,
            create_request_timeout=create_request_timeout,
            batch_size=batch_size,
        )
        return self._batch_prediction_job

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_batch_prediction_job(
        self,
        project_id: str,
        region: str,
        batch_prediction_job: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Delete a BatchPredictionJob. Can only be called on jobs that already finished.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param batch_prediction_job: The name of the BatchPredictionJob resource to be deleted.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_job_service_client(region)
        name = client.batch_prediction_job_path(project_id, region, batch_prediction_job)

        result = client.delete_batch_prediction_job(
            request={
                "name": name,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def get_batch_prediction_job(
        self,
        project_id: str,
        region: str,
        batch_prediction_job: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> BatchPredictionJob:
        """
        Get a BatchPredictionJob.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param batch_prediction_job: Required. The name of the BatchPredictionJob resource.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_job_service_client(region)
        name = client.batch_prediction_job_path(project_id, region, batch_prediction_job)

        result = client.get_batch_prediction_job(
            request={
                "name": name,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def list_batch_prediction_jobs(
        self,
        project_id: str,
        region: str,
        filter: str | None = None,
        page_size: int | None = None,
        page_token: str | None = None,
        read_mask: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> ListBatchPredictionJobsPager:
        """
        List BatchPredictionJobs in a Location.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param filter: The standard list filter.
            Supported fields:
            -  ``display_name`` supports = and !=.
            -  ``state`` supports = and !=.
            -  ``model_display_name`` supports = and !=
            Some examples of using the filter are:
            -  ``state="JOB_STATE_SUCCEEDED" AND display_name="my_job"``
            -  ``state="JOB_STATE_RUNNING" OR display_name="my_job"``
            -  ``NOT display_name="my_job"``
            -  ``state="JOB_STATE_FAILED"``
        :param page_size: The standard list page size.
        :param page_token: The standard list page token.
        :param read_mask: Mask specifying which fields to read.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_job_service_client(region)
        parent = client.common_location_path(project_id, region)

        result = client.list_batch_prediction_jobs(
            request={
                "parent": parent,
                "filter": filter,
                "page_size": page_size,
                "page_token": page_token,
                "read_mask": read_mask,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result


class BatchPredictionJobAsyncHook(GoogleBaseAsyncHook):
    """Hook for Google Cloud Vertex AI Batch Prediction Job Async APIs."""

    sync_hook_class = BatchPredictionJobHook

    def __init__(
        self,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ):
        super().__init__(
            gcp_conn_id=gcp_conn_id,
            impersonation_chain=impersonation_chain,
            **kwargs,
        )

    async def get_job_service_client(self, region: str | None = None) -> JobServiceAsyncClient:
        """Return JobServiceAsyncClient object."""
        endpoint = f"{region}-aiplatform.googleapis.com:443" if region and region != "global" else None
        return JobServiceAsyncClient(
            credentials=(await self.get_sync_hook()).get_credentials(),
            client_info=CLIENT_INFO,
            client_options=ClientOptions(api_endpoint=endpoint),
        )

    async def get_batch_prediction_job(
        self,
        project_id: str,
        location: str,
        job_id: str,
        retry: AsyncRetry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> types.BatchPredictionJob:
        """
        Retrieve a batch prediction tuning job.

        :param project_id: Required. The ID of the Google Cloud project that the job belongs to.
        :param location: Required. The ID of the Google Cloud region that the job belongs to.
        :param job_id: Required. The hyperparameter tuning job id.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client: JobServiceAsyncClient = await self.get_job_service_client(region=location)
        job_name = client.batch_prediction_job_path(project_id, location, job_id)

        result = await client.get_batch_prediction_job(
            request={
                "name": job_name,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    async def wait_batch_prediction_job(
        self,
        project_id: str,
        location: str,
        job_id: str,
        retry: AsyncRetry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        poll_interval: int = 10,
    ) -> types.BatchPredictionJob:
        statuses_complete = {
            JobState.JOB_STATE_CANCELLED,
            JobState.JOB_STATE_FAILED,
            JobState.JOB_STATE_PAUSED,
            JobState.JOB_STATE_SUCCEEDED,
        }
        while True:
            try:
                self.log.info("Requesting batch prediction tuning job with id %s", job_id)
                job: types.BatchPredictionJob = await self.get_batch_prediction_job(
                    project_id=project_id,
                    location=location,
                    job_id=job_id,
                    retry=retry,
                    timeout=timeout,
                    metadata=metadata,
                )
            except Exception as ex:
                self.log.exception("Exception occurred while requesting job %s", job_id)
                raise AirflowException(ex)

            self.log.info("Status of the batch prediction job %s is %s", job.name, job.state.name)
            if job.state in statuses_complete:
                return job

            self.log.info("Sleeping for %s seconds.", poll_interval)
            await asyncio.sleep(poll_interval)
